package netscape.palomar.sgml;

import java.io.DataOutputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import netscape.palomar.util.CascadedException;
import netscape.palomar.util.Tree;
import netscape.palomar.util.TreeImpl;

public class SGMLTagImpl extends SGMLElementImpl implements SGMLTag {
   protected String _tagType;
   protected Hashtable _attributes;
   protected Vector _attrOrder;
   protected SGMLParseTableEntry _pte;
   protected Vector _excludeList;
   protected Vector _includeList;

   public SGMLParser getParser() {
      return this._pte.parser;
   }

   public void setRawValue(String newValue) {
      super._rawValue = newValue;
      this._tagType = SGMLParser.getTagType(newValue);
   }

   public void preParseSetup(String newValue, SGMLParseTableEntry pte) {
      super._rawValue = newValue;
      this._tagType = SGMLParser.getTagType(newValue);
      this._pte = pte;
   }

   public String getTagType() {
      return this._tagType == null ? "UNKNOWN" : this._tagType;
   }

   public String getRawValue() {
      if (super._rawValue == null) {
         this.reconstructLineImage();
      }

      return super._rawValue;
   }

   void reconstructLineImage() {
      try {
         StringBuffer newImg = new StringBuffer("<");
         newImg.append(this.getTagType());
         Enumeration e = this._attrOrder.elements();

         while(e.hasMoreElements()) {
            String pName = (String)e.nextElement();
            SGMLAttribute tp = (SGMLAttribute)this._attributes.get(pName);
            if (tp != null) {
               String pVal = tp.getRawValue();
               if (pVal != null) {
                  newImg.append(" ");
                  newImg.append(pName);
                  if (!pName.equals(pVal)) {
                     newImg.append("=");
                     newImg.append(quote(pVal));
                  }
               }
            }
         }

         newImg.append(">");
         super._rawValue = newImg.toString();
      } catch (Exception e) {
         SGMLException pe = new SGMLException(215, e);
         ((CascadedException)pe).addToken("tag", this.getTagType());
         CascadedException.printException(pe, "SGMLTagImpl");
      }
   }

   public SGMLAttribute getAttribute(String name) throws CascadedException {
      try {
         if (this._attributes == null) {
            this.setUpParamVals();
         }

         return (SGMLAttribute)this._attributes.get(name.toUpperCase());
      } catch (Exception e) {
         SGMLException pe = new SGMLException(216, e);
         ((CascadedException)pe).addToken("attr", name);
         ((CascadedException)pe).addToken("tag", this.getTagType());
         throw pe;
      }
   }

   public void setAttribute(SGMLAttribute attr) throws CascadedException {
      String upperName = attr.getName().toUpperCase();

      try {
         if (this._attributes == null) {
            this.setUpParamVals();
         }

         this._attributes.put(upperName, attr);
         if (!this._attrOrder.contains(upperName)) {
            this._attrOrder.addElement(upperName);
         }

         super._rawValue = null;
      } catch (Exception e) {
         SGMLException pe = new SGMLException(217, e);
         ((CascadedException)pe).addToken("attr", upperName);
         ((CascadedException)pe).addToken("tag", this.getTagType());
         throw pe;
      }
   }

   public void initializeAttributes() throws CascadedException {
      String[] presets = this.getParser().getStandardAttributeList(this.getTagType());
      if (presets != null) {
         for(int i = 0; i < presets.length; ++i) {
            String attName = presets[i].substring(1);
            presets[i].charAt(0);
            this._attrOrder.addElement(attName);
            SGMLAttribute tp = this.createAttribute(attName, (String)null);
            this._attributes.put(attName.toUpperCase(), tp);
         }
      }

   }

   public void postProcessAttributes() throws CascadedException {
   }

   protected synchronized void setUpParamVals() throws CascadedException {
      try {
         if (this._attributes == null) {
            this._attributes = new Hashtable();
            this._attrOrder = new Vector();
            this.initializeAttributes();
            String sb = this.getRawValue();
            sb = sb.substring(1, sb.length() - 1);
            int i = this.findWhiteWithQuotes(sb, 0);
            if (i >= 0) {
               i = SGMLParser.spanWhite(sb, i);
            }

            int aStart;
            int aEnd;
            for(; i > 0; this.makeAttributeSetting(sb.substring(aStart, aEnd))) {
               aStart = i;
               aEnd = this.findWhiteWithQuotes(sb, i);
               if (aEnd < 0) {
                  this.makeAttributeSetting(sb.substring(i));
                  break;
               }

               if (sb.charAt(aEnd - 1) == '=') {
                  aEnd = SGMLParser.spanWhite(sb, aEnd);
                  aEnd = this.findWhiteWithQuotes(sb, aEnd);
                  if (aEnd < 0) {
                     this.makeAttributeSetting(sb.substring(i));
                     break;
                  }
               }

               i = SGMLParser.spanWhite(sb, aEnd);
               if (i < 0) {
                  this.makeAttributeSetting(sb.substring(aStart, aEnd));
                  break;
               }

               if (sb.charAt(i) == '=') {
                  i = SGMLParser.spanWhite(sb, i + 1);
                  aEnd = this.findWhiteWithQuotes(sb, i);
                  if (aEnd < 0) {
                     this.makeAttributeSetting(sb.substring(aStart));
                     break;
                  }

                  i = SGMLParser.spanWhite(sb, aEnd);
               }
            }

            this.postProcessAttributes();
         }
      } catch (Exception e) {
         SGMLException pe = new SGMLException(218, e);
         ((CascadedException)pe).addToken("tag", this.getTagType());
         throw pe;
      }
   }

   int findWhiteWithQuotes(String s, int start) {
      for(int i = start; i < s.length(); ++i) {
         char ch = s.charAt(i);
         if (ch <= ' ') {
            return i;
         }

         if (ch == '"') {
            i = s.indexOf(34, i + 1);
         } else if (ch == '\'') {
            i = s.indexOf(39, i + 1);
         } else if (ch == '`') {
            i = s.indexOf(96, i + 1);
         }

         if (i < 0) {
            return -1;
         }
      }

      return -1;
   }

   void makeAttributeSetting(String part) throws CascadedException {
      String key = null;

      try {
         int equalspos = part.indexOf(61);
         String value;
         if (equalspos < 0) {
            key = part.toUpperCase();
            value = key;
         } else {
            key = part.substring(0, equalspos).toUpperCase().trim();
            value = unQuote(part.substring(equalspos + 1).trim());
         }

         SGMLAttribute tp = (SGMLAttribute)this._attributes.get(key);
         if (tp == null) {
            tp = this.createAttribute(key, value);
            this._attributes.put(key, tp);
         } else {
            tp.setRawValue(value);
         }

         if (!this._attrOrder.contains(key)) {
            this._attrOrder.addElement(key);
         }
      } catch (Exception e) {
         SGMLException pe = new SGMLException(219, e);
         ((CascadedException)pe).addToken("attr", key);
         ((CascadedException)pe).addToken("tag", this.getTagType());
         throw pe;
      }
   }

   public static String unQuote(String input) throws CascadedException {
      try {
         if (input == null) {
            return null;
         } else if (input.length() < 2) {
            return input;
         } else if (input.charAt(0) == '`') {
            return input;
         } else {
            char quoteChar = input.charAt(0);
            if (quoteChar != '"' && quoteChar != '\'') {
               return input;
            } else {
               int pos = input.indexOf(quoteChar, 1);
               return pos < 0 ? input.substring(1) : input.substring(1, pos);
            }
         }
      } catch (Exception e) {
         SGMLException pe = new SGMLException(220, e);
         throw pe;
      }
   }

   public static String quote(String input) throws CascadedException {
      if (input == null) {
         return null;
      } else if (input.length() == 0) {
         return "\"\"";
      } else {
         return input.charAt(0) == '`' ? input : "\"" + input + "\"";
      }
   }

   public SGMLAttribute createAttribute(String key, String rawValue) throws CascadedException {
      SGMLAttributeImpl sai = new SGMLAttributeImpl(key, rawValue, this._pte.parser);
      return sai;
   }

   public void finishProcessing() throws CascadedException {
   }

   public Tree findGrandChildParent(Tree grandChild) {
      Tree lower = grandChild;
      Tree upper = grandChild.getParent();

      while(upper != this) {
         lower = upper;
         upper = upper.getParent();
         if (upper == null) {
            return null;
         }
      }

      return lower;
   }

   public SGMLTag parseChildren(SGMLStream inf, Vector parentExclusion, Vector parentInclusion) throws CascadedException {
      return null;
   }

   public SGMLTag parseChildrenStructure(SGMLStream inf, Vector parentExclusion, Vector parentInclusion) throws CascadedException {
      if (this._pte == null) {
         SGMLException pe = new SGMLException(221);
         ((CascadedException)pe).addToken("item", "ParseTableEntry");
         throw pe;
      } else {
         try {
            this._excludeList = (Vector)parentExclusion.clone();
            this._includeList = (Vector)parentInclusion.clone();
            int last = this._pte.exclude.size();

            for(int i1 = 0; i1 < last; ++i1) {
               this._excludeList.addElement(this._pte.exclude.elementAt(i1));
            }

            last = this._pte.include.size();

            for(int i2 = 0; i2 < last; ++i2) {
               this._includeList.addElement(this._pte.include.elementAt(i2));
            }

            String stopToken = "/" + this.getTagType();
            String sb = null;

            while(true) {
               sb = inf.readText();
               if (sb == null) {
                  return null;
               }

               if (this._pte.textOK && sb.length() > 0) {
                  SGMLText ht = inf.generateText(sb);
                  ((TreeImpl)this).addChildQuiet(ht);
               }

               sb = inf.readTag();
               if (sb == null) {
                  return null;
               }

               SGMLTag returnedtx;
               if (sb.length() > 0) {
                  for(SGMLTag tx = inf.generateTag(sb); tx != null; tx = returnedtx) {
                     String txtype = tx.getTagType();
                     if (txtype.equals(stopToken)) {
                        return null;
                     }

                     if (!this.allowedChildType(txtype) && this.allowedInParent(txtype)) {
                        return tx;
                     }

                     this.appendParsedChild(tx);
                     returnedtx = tx.parseChildren(inf, this._excludeList, this._includeList);

                     try {
                        tx.finishProcessing();
                     } catch (Exception ex) {
                        SGMLException se = new SGMLException(227, ex);
                        ((CascadedException)se).addToken("tag", this.getTagType());
                        ((CascadedException)se).addToken("line", Integer.toString(inf.getLineNumber()));
                        ((CascadedException)se).addToken("col", Integer.toString(inf.getColumnNumber()));
                        inf.registerWarning(se);
                     }
                  }
               }
            }
         } catch (Exception e) {
            SGMLException pe = new SGMLException(222, e);
            ((CascadedException)pe).addToken("tag", this.getTagType());
            throw pe;
         }
      }
   }

   public SGMLTag parseChildrenTop(SGMLStream inf, Vector parentExclusion, Vector parentInclusion) throws CascadedException {
      try {
         String sb = null;
         this._excludeList = parentExclusion;
         this._includeList = parentInclusion;

         while(true) {
            sb = inf.readText();
            if (sb == null) {
               return null;
            }

            if (sb.equals("\n")) {
               sb = "";
            }

            if (sb.length() > 0) {
               SGMLText ht = inf.generateText(sb);
               ((TreeImpl)this).addChildQuiet(ht);
            }

            sb = inf.readTag();
            if (sb == null) {
               return null;
            }

            SGMLTag returnedtx;
            if (sb.length() > 0) {
               for(SGMLTag tx = inf.generateTag(sb); tx != null; tx = returnedtx) {
                  tx.getTagType();
                  ((TreeImpl)this).addChildQuiet(tx);
                  returnedtx = tx.parseChildren(inf, parentExclusion, parentInclusion);

                  try {
                     tx.finishProcessing();
                  } catch (Exception ex) {
                     SGMLException se = new SGMLException(227, ex);
                     ((CascadedException)se).addToken("tag", this.getTagType());
                     ((CascadedException)se).addToken("line", Integer.toString(inf.getLineNumber()));
                     ((CascadedException)se).addToken("col", Integer.toString(inf.getColumnNumber()));
                     inf.registerWarning(se);
                  }
               }
            }
         }
      } catch (Exception e) {
         SGMLException pe = new SGMLException(222, e);
         ((CascadedException)pe).addToken("tag", "Top");
         throw pe;
      }
   }

   public SGMLTag parseChildrenScript(SGMLStream inf, Vector parentExclusion, Vector parentInclusion) throws CascadedException {
      if (this._pte == null) {
         SGMLException pe = new SGMLException(221);
         ((CascadedException)pe).addToken("item", "ParseTableEntry");
         throw pe;
      } else {
         this._excludeList = parentExclusion;
         this._includeList = parentInclusion;
         String stopToken = "/" + this.getTagType();
         String sb = null;

         while(inf.textNext()) {
            sb = inf.getTextTilToken(stopToken);
            if (sb == null) {
               return null;
            }

            SGMLText ht = inf.generateScript(sb);
            ((TreeImpl)this).addChildQuiet(ht);
         }

         sb = inf.readTag();
         return null;
      }
   }

   public void appendParsedChild(SGMLTag tx) throws CascadedException {
      ((TreeImpl)this).addChildQuiet(tx);
   }

   public boolean allowedChildType(String testTag) {
      boolean allowed = this._pte.contentModel.contains(testTag);
      if (this._includeList == null) {
         return allowed;
      } else {
         if (!allowed) {
            allowed = this._includeList.contains(testTag);
            if (!allowed) {
               if (testTag.length() < 1) {
                  return false;
               }

               if (testTag.charAt(0) == '/') {
                  String stopTag = "/" + this.getTagType();
                  if (stopTag.equals(testTag)) {
                     return true;
                  }
               }

               return false;
            }
         }

         return !this._excludeList.contains(testTag);
      }
   }

   public boolean allowedInParent(String testTag) throws CascadedException {
      Tree parent = ((TreeImpl)this).getParent();

      while(parent != null) {
         if (parent instanceof SGMLTag) {
            if (((SGMLTag)parent).allowedChildType(testTag)) {
               if (((SGMLTag)parent).isTopTag()) {
                  return false;
               }

               return true;
            }

            parent = parent.getParent();
         }
      }

      return false;
   }

   public void allowableChild(SGMLElement newTree) throws CascadedException {
      if (newTree instanceof SGMLText) {
         if (!this._pte.textOK) {
            SGMLException pe = new SGMLException(223);

            try {
               ((CascadedException)pe).addToken("tag", "'text'");
               ((CascadedException)pe).addToken("parent", this.getTagType());
            } catch (Exception var3) {
            }

            throw pe;
         }
      } else if (!this.allowedChildType(newTree.getTagType())) {
         SGMLException pe = new SGMLException(223);

         try {
            ((CascadedException)pe).addToken("tag", newTree.getTagType());
            ((CascadedException)pe).addToken("parent", this.getTagType());
         } catch (Exception var4) {
         }

         throw pe;
      } else {
         if (newTree instanceof SGMLTag) {
            ((SGMLTag)newTree).checkContext(this._excludeList, this._includeList);
         }

      }
   }

   public void checkContext(Vector parentExclusion, Vector parentInclusion) throws CascadedException {
      this._excludeList = (Vector)parentExclusion.clone();
      Enumeration e = this._pte.exclude.elements();

      while(e.hasMoreElements()) {
         this._excludeList.addElement(e.nextElement());
      }

      this._includeList = (Vector)parentInclusion.clone();
      Enumeration e = this._pte.exclude.elements();

      while(e.hasMoreElements()) {
         this._excludeList.addElement(e.nextElement());
      }

      Enumeration e = ((TreeImpl)this).getChildren();

      while(e.hasMoreElements()) {
         SGMLElement testChild = (SGMLElement)e.nextElement();
         if (testChild instanceof SGMLTag && !this.allowedChildType(testChild.getTagType())) {
            SGMLException pe = new SGMLException(224);
            ((CascadedException)pe).addToken("parent", this.getTagType());
            ((CascadedException)pe).addToken("child", testChild.getTagType());
            throw pe;
         }

         if (testChild instanceof SGMLTag) {
            ((SGMLTag)testChild).checkContext(this._excludeList, this._includeList);
         }
      }

   }

   public void writeYourself(DataOutputStream dos) throws CascadedException {
      try {
         dos.writeBytes(this.getRawValue());
         if (!this._pte.contentEmpty) {
            boolean addReturns = !this._pte.textOK;
            int last = ((TreeImpl)this).numChildren();

            for(int i = 0; i < last; ++i) {
               Tree child = ((TreeImpl)this).childAt(i);
               if (child instanceof SGMLElement) {
                  ((SGMLElement)child).writeYourself(dos);
               }
            }

            this.writeVirtualContent(dos);
            if (addReturns) {
               dos.writeBytes("\n");
            }

            dos.writeBytes("</");
            dos.writeBytes(this.getTagType());
            dos.writeBytes(">");
         }
      } catch (Exception e) {
         SGMLException ce = new SGMLException(225, e);
         throw ce;
      }
   }

   public void writeVirtualContent(DataOutputStream dos) throws CascadedException {
   }

   public boolean isTopTag() {
      return false;
   }

   public void beenCloned() throws CascadedException {
      this._excludeList = null;
      this._includeList = null;
      Hashtable oldAttrs = this._attributes;
      super.beenCloned();

      try {
         if (this._attributes != null) {
            this._attrOrder = (Vector)this._attrOrder.clone();
            this._attributes = new Hashtable();
            Enumeration eKeys = oldAttrs.keys();

            while(eKeys.hasMoreElements()) {
               Object oKey = eKeys.nextElement();
               SGMLAttribute tp = (SGMLAttribute)oldAttrs.get(oKey);
               SGMLAttribute tpclone = this.createAttribute(tp.getName(), tp.getRawValue());
               this._attributes.put(oKey, tpclone);
            }

         }
      } catch (Exception e) {
         SGMLException pe = new SGMLException(226, e);
         ((CascadedException)pe).addToken("class", this.getTagType());
         ((CascadedException)pe).addToken("name", this.getTagType());
         throw pe;
      }
   }

   public void flush() throws CascadedException {
      this._attributes = null;
      this._attrOrder = null;
      this._excludeList = null;
      this._includeList = null;
      super.flush();
   }
}
